@@ -1,7 +1,7 @@
 package Perinci::CmdLine::Inline;
 
-our $DATE = '2015-09-28'; # DATE
-our $VERSION = '0.24'; # VERSION
+our $DATE = '2015-09-29'; # DATE
+our $VERSION = '0.25'; # VERSION
 
 use 5.010001;
 use strict;
@@ -132,9 +132,7 @@ might also want to try using `fatten` to fatpack your
 
 TODO:
 
-* Option to validate argument (using periswrap).
-
-* Support per_arg_json.
+* Option to validate argument (embedding code generated by periswrap).
 
 _
     args_rels => {
@@ -288,6 +286,7 @@ sub gen_inline_pericmd_script {
                 "Data::Check::Structure",
                 "Text::Table::Tiny",
                 "Getopt::Long::EvenLess",
+                "Perinci::Result::Format::Lite",
                 @{ $args{include} // [] },
             )
         },
@@ -350,16 +349,6 @@ _
     print "\n";
 _
 
-        # from List::MoreUtils 0.401
-        $cd->{subs}{_pci_firstidx} = ['&@', <<'_'];
-    my $f = shift;
-    foreach my $i ( 0 .. $#_ )
-    {
-        local *_ = \$_[$i];
-        return $i if $f->();
-    }
-    return -1;
-_
         $cd->{subs}{_pci_json} = <<'_';
      state $json = do {
         # XXX try JSON::XS first, fallback to JSON::PP
@@ -376,167 +365,11 @@ _
             $cd->{module_srcs}{'Inlined::_pci_clean_json'} = "require Scalar::Util; use feature 'state'; sub _pci_clean_json { $src }\n1;\n";
         }
 
-        # borrowed from Perinci::CmdLine::Lite 1.13, use our own firstidx
-        $cd->{module_srcs}{'Inlined::_pci_gen_table'} = <<'_';
-sub _pci_gen_table {
-    my ($data, $header_row, $resmeta, $is_pretty) = @_;
-
-    $resmeta //= {};
-
-    my @columns;
-    if ($header_row) {
-        @columns = @{$data->[0]};
-    } else {
-        @columns = map {"col$_"} 0..@{$data->[0]}-1;
-    }
-
-    my $column_orders; # e.g. [col2, col1, col3, ...]
-  SET_COLUMN_ORDERS: {
-
-        # find column orders from 'table_column_orders' in result metadata (or
-        # from env)
-        my $tcos;
-        if ($ENV{FORMAT_PRETTY_TABLE_COLUMN_ORDERS}) {
-            $tcos = _pci_json()->decode($ENV{FORMAT_PRETTY_TABLE_COLUMN_ORDERS});
-        } elsif (my $rfos = ($resmeta->{'cmdline.format_options'} //
-                                 $resmeta->{format_options})) {
-            my $rfo = $rfos->{'text-pretty'} // $rfos->{text} // $rfos->{any};
-            if ($rfo) {
-                $tcos = $rfo->{table_column_orders};
-            }
-        }
-        if ($tcos) {
-            # find an entry in tcos that @columns contains all the columns of
-          COLS:
-            for my $cols (@$tcos) {
-                for my $col (@$cols) {
-                    next COLS unless first {$_ eq $col} @columns;
-                }
-                $column_orders = $cols;
-                last SET_COLUMN_ORDERS;
-            }
-        }
-
-        # find column orders from table spec
-        $column_orders = $resmeta->{'table.fields'};
-    }
-
-    # reorder each row according to requested column order
-    if ($column_orders) {
-        # 0->2, 1->0, ... (map column position from unordered to ordered)
-        my @map0 = sort {
-            my $idx_a = _pci_firstidx(sub {$_ eq $a->[1]},
-                                                  @$column_orders) // 9999;
-            my $idx_b = _pci_firstidx(sub {$_ eq $b->[1]},
-                                                  @$column_orders) // 9999;
-            $idx_a <=> $idx_b || $a->[1] cmp $b->[1];
-        } map {[$_, $columns[$_]]} 0..@columns-1;
-        #use DD; dd \@map0;
-        my @map;
-        for (0..@map0-1) {
-            $map[$_] = $map0[$_][0];
-        }
-        #use DD; dd \@map;
-        my $newdata = [];
-        for my $row (@$data) {
-            my @newrow;
-            for (0..@map-1) { $newrow[$_] = $row->[$map[$_]] }
-            push @$newdata, \@newrow;
-        }
-        $data = $newdata;
-    }
-
-    if ($is_pretty) {
-        require Text::Table::Tiny;
-        Text::Table::Tiny::table(rows=>$data, header_row=>$header_row) . "\n";
-    } else {
-        no warnings 'uninitialized';
-        shift @$data if $header_row;
-        join("", map {join("\t", @$_)."\n"} @$data);
-    }
-}
-1;
-_
-
-$cd->{module_srcs}{'Inlined::_pci_format_result'} = <<'_';
-sub _pci_format_result {
-    require Data::Check::Structure;
-
-    my $r = shift;
-
-    my $res    = $r->{res};
-    my $format = $r->{format} // 'text';
-
-    if ($format =~ /\Atext(-simple|-pretty)?\z/) {
-        my $is_pretty = $format eq 'text-pretty' ? 1 :
-            $format eq 'text-simple' ? 0 : (-t STDOUT);
-        no warnings 'uninitialized';
-        if ($res->[0] !~ /^(2|304)/) {
-            my $fres = "ERROR $res->[0]: $res->[1]";
-            if (my $prev = $res->[3]{prev}) {
-                $fres .= " ($prev->[0]: $prev->[1])";
-            }
-            return "$fres\n";
-        } elsif ($res->[3] && $res->[3]{"x.hint.result_binary"}) {
-            return $res->[2];
-        } else {
-            my $data = $res->[2];
-            my $max = 5;
-            if (!ref($data)) {
-                $data //= "";
-                $data .= "\n" unless !length($data) || $data =~ /\n\z/;
-                return $data;
-            } elsif (ref($data) eq 'ARRAY' && !@$data) {
-                return "";
-            } elsif (Data::Check::Structure::is_aos($data, {max=>$max})) {
-                return join("", map {"$_\n"} @$data);
-            } elsif (Data::Check::Structure::is_aoaos($data, {max=>$max})) {
-                require Inlined::_pci_gen_table;
-                return _pci_gen_table($data, 0, $res->[3], $is_pretty);
-            } elsif (Data::Check::Structure::is_hos($data, {max=>$max})) {
-                $data = [map {[$_, $data->{$_}]} sort keys %$data];
-                unshift @$data, ["key", "value"];
-                require Inlined::_pci_gen_table;
-                return _pci_gen_table($data, 1, $res->[3], $is_pretty);
-            } elsif (Data::Check::Structure::is_aohos($data, {max=>$max})) {
-                # collect all mentioned fields
-                my %fieldnames;
-                for my $row (@$data) {
-                    $fieldnames{$_}++ for keys %$row;
-                }
-                my @fieldnames = sort keys %fieldnames;
-                my $newdata = [];
-                for my $row (@$data) {
-                    push @$newdata, [map {$row->{$_}} @fieldnames];
-                }
-                unshift @$newdata, \@fieldnames;
-                require Inlined::_pci_gen_table;
-                return _pci_gen_table($newdata, 1, $res->[3], $is_pretty);
-            } else {
-                $format = 'json-pretty';
-            }
-        }
-    }
-
-    $res = $res->[2] if $r->{naked_res};
-
-    warn "Unknown format '$format', fallback to json-pretty"
-        unless $format =~ /\Ajson(-pretty)?\z/;
-    require Inlined::_pci_clean_json;
-    _pci_clean_json($res);
-    if ($format eq 'json') {
-        return _pci_json()->encode($res) . "\n";
-    } else {
-        return _pci_json()->canonical(1)->pretty->encode($res);
-    }
-}
-1;
-_
-
         require Perinci::Sub::GetArgs::Argv;
         my $ggl_res = Perinci::Sub::GetArgs::Argv::gen_getopt_long_spec_from_meta(
             meta => $meta,
             meta_is_normalized => 1,
+            per_arg_json => 1,
             common_opts => \%copts,
         );
         return [500, "Can't generate Getopt::Long spec from meta: ".
@@ -603,7 +436,7 @@ _
         }
 
         $cd->{vars}{'$_pci_r'}++;
-        push @l, '$_pci_r = { format=>"text", naked_res=>0, };', "\n\n";
+        push @l, '$_pci_r = { naked_res=>0, };', "\n\n";
 
         # gen code to parse cmdline options
         $cd->{vars}{'%_pci_args'}++;
@@ -664,6 +497,8 @@ _
                     } else {
                         if (($specmeta->{parsed}{type} // '') =~ /\@/) {
                             push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
+                        } elsif ($specmeta->{is_json}) {
+                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json->decode($_[1]);';
                         } else {
                             push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
                         }
@@ -697,15 +532,18 @@ _
         push @l, "# display result\n\n";
         push @l, "{\n";
         push @l, 'my $fres;', "\n";
-        push @l, 'if (', ($skip_format ? 1:0), ' || $_pci_r->{res}[3]{"cmdline.skip_format"}) { $fres = $_pci_r->{res}[2] } else { require Inlined::_pci_format_result; $fres = _pci_format_result($_pci_r) }', "\n";
+        push @l, 'my $save_res; if (exists $_pci_r->{res}[3]{"cmdline.result"}) { $save_res = $_pci_r->{res}[2]; $_pci_r->{res}[2] = $_pci_r->{res}[3]{"cmdline.result"} }', "\n";
+        push @l, 'if (', ($skip_format ? 1:0), ' || $_pci_r->{res}[3]{"cmdline.skip_format"}) { $fres = $_pci_r->{res}[2] } else { require Perinci::Result::Format::Lite; $fres = Perinci::Result::Format::Lite::format($_pci_r->{res}, ($_pci_r->{format} // $_pci_r->{res}[3]{"cmdline.default_format"} // "text"), $_pci_r->{naked_res}, 0) }', "\n";
         push @l, 'print $fres;', "\n";
+        push @l, 'if (defined $save_res) { $_pci_r->{res}[2] = $save_res }', "\n";
         push @l, "}\n\n";
 
         # generate code to exit with code
         push @l, "# exit\n\n";
         push @l, "{\n";
         push @l, 'my $status = $_pci_r->{res}[0];', "\n";
-        push @l, 'exit($status =~ /200|304/ ? 0 : ($status-300));', "\n";
+        push @l, 'my $exit_code = $_pci_r->{res}[3]{"cmdline.exit_code"} // ($status =~ /200|304/ ? 0 : ($status-300));', "\n";
+        push @l, 'exit($exit_code);', "\n";
         push @l, "}\n\n";
 
         if ($args{pass_cmdline_object}) {
@@ -842,7 +680,7 @@ Perinci::CmdLine::Inline - Generate inline Perinci::CmdLine CLI script
 
 =head1 VERSION
 
-This document describes version 0.24 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2015-09-28.
+This document describes version 0.25 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2015-09-29.
 
 =head1 SYNOPSIS
 
@@ -868,23 +706,23 @@ B<EARLY DEVELOPMENT.>
 
 Sample startup overhead benchmark vs Perinci::CmdLine::Lite:
 
-                                   Rate  run 1 (lite) run --version (lite)    -c (lite) run 1 (inline) run --version (inline) -c (inline)   perl
- run 1 (lite)           12.931+-0.013/s            --               -60.8%       -83.1%         -89.2%                 -91.5%      -93.8% -97.6%
- run --version (lite)   33.016+-0.057/s 155.33+-0.51%                   --       -56.8%         -72.3%                 -78.3%      -84.1% -93.9%
- -c (lite)                76.46+-0.24/s     491.3+-2%         131.6+-0.84%           --         -35.9%                 -49.8%      -63.2% -85.9%
- run 1 (inline)          119.27+-0.44/s   822.4+-3.5%          261.3+-1.5% 55.99+-0.76%             --                 -21.7%      -42.6% -78.1%
- run --version (inline)  152.26+-0.51/s  1077.5+-4.1%          361.2+-1.7% 99.12+-0.92%   27.65+-0.63%                     --      -26.7% -72.0%
- -c (inline)             207.68+-0.88/s    1506.1+-7%            529+-2.9%  171.6+-1.4%   74.12+-0.97%            36.4+-0.73%          -- -61.8%
- perl                        544+-2.5/s     4107+-20%         1547.6+-8.2%    611.4+-4%    356.1+-2.7%              257.3+-2% 161.9+-1.6%     --
+                                   Rate run 1 (lite) run --version (lite)    -c (lite) run 1 (inline) run --version (inline) -c (inline)   perl
+ run 1 (lite)            12.75+-0.011/s           --               -60.2%       -82.9%         -87.7%                 -91.4%      -93.6% -97.7%
+ run --version (lite)   32.067+-0.096/s 151.5+-0.78%                   --       -57.0%         -69.0%                 -78.3%      -83.8% -94.1%
+ -c (lite)                74.65+-0.18/s  485.4+-1.5%         132.78+-0.9%           --         -27.9%                 -49.4%      -62.4% -86.3%
+ run 1 (inline)          103.58+-0.33/s  712.4+-2.7%            223+-1.4% 38.76+-0.56%             --                 -29.8%      -47.8% -80.9%
+ run --version (inline)      147.6+-1/s 1057.9+-8.2%          360.4+-3.5%   97.8+-1.5%     42.5+-1.1%                     --      -25.6% -72.8%
+ -c (inline)               198.4+-1.3/s 1456.3+-9.9%          518.8+-4.3%  165.8+-1.8%     91.6+-1.4%             34.4+-1.3%          -- -63.5%
+ perl                      543.3+-3.6/s    4161+-28%            1594+-12%  627.8+-5.1%    424.5+-3.8%              268+-3.5% 173.8+-2.5%     --
  
  Average times:
-   perl                  :     1.8382ms
-   -c (inline)           :     4.8151ms
-   run --version (inline):     6.5677ms
-   run 1 (inline)        :     8.3843ms
-   -c (lite)             :    13.0787ms
-   run --version (lite)  :    30.2883ms
-   run 1 (lite)          :    77.3335ms
+   perl                  :     1.8406ms
+   -c (inline)           :     5.0403ms
+   run --version (inline):     6.7751ms
+   run 1 (inline)        :     9.6544ms
+   -c (lite)             :    13.3958ms
+   run --version (lite)  :    31.1847ms
+   run 1 (lite)          :    78.4314ms
 
 =head1 FUNCTIONS
 
@@ -938,12 +776,12 @@ TODO:
 
 =over
 
-=item * Option to validate argument (using periswrap).
-
-=item * Support per_arg_json.
+=item * Option to validate argument (embedding code generated by periswrap).
 
 =back
 
+This function is not exportable.
+
 Arguments ('*' denotes required arguments):
 
 =over 4
